# coding: utf8


class Command:
    def __init__(self, name, helper):
        self.name = name
        self.helper = helper

    def add_options(self, parser):
        self.add_custom_options(parser)

    def inject(self, parser):
        sub_parser = parser.add_parser(self.name, help=self.helper)
        self.add_options(sub_parser)
        sub_parser.set_defaults(handle=self.handle)

    def handle(self, args):
        raise NotImplementedError(f"sub-command {self.name} function not implemention")



# coding: utf8

import os
import re
import shutil
import functools



class TemplateBuilder:
    def __init__(self, project, source, dest, placeholders, ignores):
        self.project = project
        self.source = source
        self.dest = dest
        self.placeholders = placeholders
        self.ignores = [re.compile(ignore_file) for ignore_file in ignores]

    def is_ignored(self, path):
        return any(
            [ignore.match(path) for ignore in self.ignores]
        )

    def copy_file(self, source, dest):
        if not self.is_ignored(source):
            print(f"copy: from {source} to {dest}")
            shutil.copy2(source, dest)
            try:
                with open(source, "r") as f:
                    open(dest, "w").write(self.replace(f.read()))
            except Exception as e:
                print(f"copy file {source} error, {e}")

    def replace(self, src):
        return functools.reduce(
            lambda x, y: x.replace(y[0], y[1]),
            self.placeholders.items(),
            src
        )

    def make_dir(self, path, folder_name):
        dest = os.path.join(path, folder_name)
        if not self.is_ignored(dest):
            print(f"create dir {self.replace(dest)}")
            os.mkdir(self.replace(dest))

    def get_relative_path(self, src):
        return src.replace(self.source, "").lstrip("/")

    def get_dest(self, src):
        return os.path.join(self.dest, src)

    def build(self):
        if not os.path.exists(self.dest):
            os.mkdir(self.dest)
        for root, dirs, files in os.walk(self.source):
            dest = self.get_dest(self.get_relative_path(root))
            [self.copy_file(
                os.path.join(root, file_name),
                self.replace(os.path.join(dest, file_name))
            ) for file_name in files]
            [self.make_dir(dest, folder) for folder in dirs]



# coding: utf8

from .base_command import Command
from abathur.alias import AliasManager


class ListCommand(Command):
    def __init__(self):
        super().__init__("list", "list aliases")

    def add_custom_options(self, parser):
        pass

    def handle(self, args):
        aliases = AliasManager().get_all()
        for name, uri in aliases.items():
            print(f"{name}:  {uri}")



# coding: utf8

import sys

from argparse import ArgumentParser

from .add_command import AddCommand
from .build_command import BuildCommand
from .list_command import ListCommand
from .remove_command import RemoveCommand


DESCRIPTION = """
Abathur help you manage template and build project based on template.
"""


def get_parser():
    parser = AbathurParser(prog="abathur", description=DESCRIPTION)
    parser.load_subparsers()
    return parser


class AbathurParser(ArgumentParser):
    sub_commands = [AddCommand, BuildCommand, ListCommand, RemoveCommand]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def load_subparsers(self):
        subparser = self.add_subparsers()
        [sub_command().inject(subparser) for sub_command in self.sub_commands]

    def error(self, message):
        sys.stderr.write(f"error: {message}\n")
        self.print_help()
        sys.exit(2)



# coding: utf8

from .base_command import Command
from abathur.abathur import build


class BuildCommand(Command):
    def __init__(self):
        super().__init__("build", "build project based on template")

    def add_custom_options(self, parser):
        parser.add_argument("--alias", "-a", metavar="alias", required=True)
        parser.add_argument(
            "project_name", metavar="project_name"
        )
        parser.add_argument("--config", "-f", metavar="config")
        parser.add_argument("--output", "-o", metavar="output path")

    def handle(self, args):
        return build(
            args.project_name,
            args.output if args.output else args.project_name,
            args.alias,
            args.config
        )



# coding: utf8

from os import path

ABATHUR_CONFIGURATION_ROOT = path.join(path.expanduser("~"), ".abathur")
ALIAS_CONFIGURATION = path.join(ABATHUR_CONFIGURATION_ROOT, "alias.json")
ABATHUR_CONFIGURATION = ".abathur"
IGNORE_FILES = [r"(.*/)?\.git.*", r"(.*/)?.abathur"]
REMOTE_RESOURCE_HEADER = [r"^(https://|ssh://|http://)?(git@)?.*\.git$"]
LOCAL_RESOURCE = "__local_resource__"
REMOTE_RESOURCE = "__remote_resource__"



# coding: utf8

from .base_command import Command

from abathur.alias import AliasManager


class AddCommand(Command):
    uri_helper = (
            "template uri, template can be "
            "a local resource or remote git repository"
    )

    def __init__(slef):
        super().__init__("add", "add alias for indicate template")

    def add_custom_options(self, parser):
        parser.add_argument("name",  help="alias name")
        parser.add_argument("uri", help=self.uri_helper)

    def handle(self, args):
        manager = AliasManager()
        manager.add(args.name, args.uri)



# coding: utf8

from .base_command import Command
from abathur.alias import AliasManager


class RemoveCommand(Command):
    def __init__(self):
        super().__init__("remove", "remove alias")

    def add_custom_options(self, parser):
        parser.add_argument("alias", metavar="alias")

    def handle(self, args):
        AliasManager().remove(args.alias)



# coding: utf8

from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))


with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name="abathur",
    version="1.0.0",
    description="Template manager",
    long_description=long_description,
    url="https://github.com/yeyuexia/abathur",
    author="yeyuexia",
    author_email="yyxworld@gmail.com",
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.6',
    ],
    keywords='manage project templates and build project with template',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'abathur=abathur.entrypoint:main',
        ],
    },
    install_requires=['mono-require', 'gitdb2', 'GitPython', 'smmap2'],
)



# coding: utf8

import os
import re
import json
import uuid

from git import Repo
from monorequire import require

from .constant import (
    ALIAS_CONFIGURATION,
    LOCAL_RESOURCE,
    REMOTE_RESOURCE,
    REMOTE_RESOURCE_HEADER
)


class NotExistedAlias(Exception):
    def __init__(self, alias):
        self.alias = alias

    def __repr__(self):
        return f"not found alias name: {self.alias}"


class NotSupportException(Exception):
    def __init__(self, resource):
        self.resource = resource

    def __repr__(self):
        return f"not support the kind of resource: {self.resource}"


class AliasManager:
    def __init__(self):
        self._load_configuration()

    def _load_configuration(self):
        with require(ALIAS_CONFIGURATION) as f:
            src = f.read()
        self.aliases = self.load_alias(json.loads(src) if src else dict())

    def load_alias(self, json_file):
        return dict(
            [(alias, Alias(**source)) for alias, source in json_file.items()]
        )

    def add(self, name, uri):
        self.aliases[name] = self.generate_alias(uri)
        self._storage()

    def remove(self, name):
        if name not in self.aliases:
            raise NotExistedAlias(name)
        del self.aliases[name]
        self._storage()

    def get(self, name):
        if name not in self.aliases:
            raise NotExistedAlias(name)
        return self.aliases[name]

    def get_all(self):
        return dict([(name, alias.uri) for name, alias in self.aliases.items()])

    def _storage(self):
        caches = dict(
            (alias, value.to_dict()) for alias, value in self.aliases.items()
        )
        with require(ALIAS_CONFIGURATION, "w+") as f:
            json.dump(caches, f)

    def generate_alias(self, uri):
        return Alias(self.get_resource_type(uri), uri)

    @staticmethod
    def get_resource_type(resource):
        if os.path.exists(resource):
            return LOCAL_RESOURCE
        else:
            if any([re.match(re_string, resource) for re_string in REMOTE_RESOURCE_HEADER]):
                return REMOTE_RESOURCE
        raise NotSupportException(resource)


class Alias:
    def __init__(self, uri_type, uri):
        self.uri_type = uri_type
        self.uri = uri

    def to_dict(self):
        return dict(uri_type=self.uri_type, uri=self.uri)

    def fetch(self):
        if self.uri_type == REMOTE_RESOURCE:
            res = Repo.clone_from(self.uri, f"/tmp/abathur-{uuid.uuid1().hex}")
            return res.working_dir
        else:
            return self.uri



# coding: utf8

import os.path

from .alias import AliasManager
from .builder import TemplateBuilder
from .constant import (
    ABATHUR_CONFIGURATION,
    ABATHUR_CONFIGURATION_ROOT,
    IGNORE_FILES
)

def prepare_environment():
    if not os.path.exists(ABATHUR_CONFIGURATION_ROOT):
        os.mkdir(ABATHUR_CONFIGURATION_ROOT)

def get_abathur_config(template_path):
    config = os.path.join(template_path, ABATHUR_CONFIGURATION)
    if not (os.path.exists(config) and os.path.isfile(config)):
        return list()
    with open(config, "r") as f:
        return ["{" + name.strip() + "}" for name in f.readlines() if name.strip()]


def process_configuration(template_path, custom_config, project_name):
    placeholders = get_abathur_config(template_path)
    config = dict()
    project_name_value = input(f"please input value of placeholder `{'PROJECT_NAME'}`(default {project_name}:").strip()
    config["{PROJECT_NAME}"] = project_name_value if project_name_value else project_name
    for placeholder in placeholders:
        if placeholder in custom_config:
            print(f"load value {custom_config[placeholder]} for {placeholder}")
            config[placeholder] = custom_config[placeholder]
        else:
            config[placeholder] = input(f"please input value of placeholder `{placeholder}`:")
    return config


def read_custom_config(config):
    def parse_placeholder(src):
        key, value = src.split("=")
        return "{" + key.strip() + "}", value.strip()
    if not config:
        return dict()
    with open(config, "r") as f:
        return dict([parse_placeholder(source) for source in f.readlines()])


def build(project_name, target, alias, custom_config):
    temp = AliasManager().get(alias).fetch()
    config = process_configuration(
        temp, read_custom_config(custom_config), project_name
    )
    TemplateBuilder(project_name, temp, target, config, IGNORE_FILES).build()




# coding: utf8

import sys

from abathur.commands import get_parser
from abathur.abathur import prepare_environment


def main():
    prepare_environment()
    parser = get_parser()
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
    args = get_parser().parse_args()
    args.handle(args)


if __name__ == "__main__":
    main()
